#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Applicable to webstarts"""

import logging

from celery import signals
from kids.cache import cache
from raven.contrib.celery import CeleryFilter, SentryCeleryHandler
from structlog import get_logger

from . import defaults, log_id, trace, wflask

__author__ = "john"

logger = get_logger(__name__)


def setup_celery(_):
  """Register celery shit"""

  signals.setup_logging.connect(on_setup)
  signals.before_task_publish.connect(on_before_pub)
  signals.task_prerun.connect(on_prerun)
  get_sentry_handler().install()
  add_celery_filter()

  logger.info('Configured Celery')
  set_logs_to_info()


def on_setup(**_):
  pass


@cache(key=defaults.CACHE_KEY)
def get_sentry_handler():
  return SentryCeleryHandler(wflask.get_client())


def add_celery_filter():
  handler = wflask.sentry_handler()
  if not [f for f in handler.filters if isinstance(f, CeleryFilter)]:
    handler.addFilter(CeleryFilter())


@trace
def on_before_pub(headers=None, **_):
  headers[defaults.LOG_KEY] = log_id.find()


@trace
def on_prerun(task_id=None, task=None, **_):
  logid = getattr(task.request, 'logid', None)
  # ctx = dict(tid=task_id, task=task.name, logid=logid)
  ctx = dict(logid=logid)
  logger.new(**ctx)


def set_logs_to_info():
  names = [
    'amqp',
    'kombu',
    'celery'
  ]
  [logging.getLogger(n).setLevel(logging.INFO) for n in names]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Applicable to webstarts"""
import logging
from gunicorn.app.wsgiapp import WSGIApplication
from gunicorn.glogging import Logger

__author__ = "john"


from . import types

class WebstartsApp(WSGIApplication):
  """Gunicorn wrapper"""

  def __init__(self, program, is_dev=False):
    self.is_dev = is_dev
    super().__init__(program)

  def load_wsgiapp(self):
    app = super().load_wsgiapp()
    from . import log_id, wflask, util

    app = util.wrap_session(app)
    app = log_id.wrap_logid(app)
    app = wflask.wrap_sentry(app)
    return app

  def init(self, parser, opts, args):
    """Set debug settings if needed"""

    cfg = super().init(parser, opts, args) or {}
    cfg['logger_class'] = GunicornLogger
    if self.is_dev:
      cfg.update(reload=True, timeout=99999, accesslog='-')
    return cfg


class GunicornLogger(Logger):
  def setup(self, cfg):
    """remove default error handler"""
    super().setup(cfg)
    self.error_log.propagate = True
    self._set_handler(self.error_log, None, None)
    self.access_log.propagate = True
    self._set_handler(self.access_log, None, None)

  @classmethod
  def install(cls):
    logging.setLoggerClass(types.WebstartsLogger)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Applicable to webstarts"""
import logging

from flask import Flask
from kids.cache import cache
from raven import Client, middleware
from raven.contrib.flask import Sentry
from raven.handlers.logging import SentryHandler

from . import defaults, trace

__author__ = "john"


def setup_app(app, dsn=None, **client_args):
  """

  - exclude_paths
    Should use this to ignore the flask-classful crap that is in every trace?
  - Pass all in 'data'. logging.error('asdf', extra={'data': {}})
  - When updating sentry with app
    - sentry.client.tags['flask_app'] = app.name
  """
  sentry = get_sentry()
  for k, v in client_args.items():
    setattr(sentry.client, k, v)
  sentry.init_app(app, dsn=dsn)
  sentry.client.tags['flask_app'] = app.name
  logging.getLogger(__name__).info('Finalized webstarts config')


class FlaskApp(Flask):
  """Inherit from this in usage application"""

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  @property
  def logger(self):
    return logging.getLogger(self.logger_name)


@cache(key=defaults.CACHE_KEY)
def get_sentry(**kwargs) -> 'Sentry':
  kwargs['install_logging_hook'] = False
  logger = logging.getLogger(__name__)
  # ignore not configured message from raven that is forced on us at info
  _log = logging.getLogger('webstarts.wflask.SentryClient')
  _log.setLevel(logging.WARN)
  client = SentryClient(**kwargs)
  _log.setLevel(logging.DEBUG)
  sentry = Sentry(client=client, wrap_wsgi=False, logging=False)
  get_sentry.sentry = sentry

  logger.info('Configured Raven')
  return get_sentry.sentry


def get_client() -> 'SentryClient':
  return get_sentry().client


@cache(key=defaults.CACHE_KEY)
def sentry_handler() -> 'SentryHandler':
  sentry = get_sentry()
  handler = SentryHandler(client=sentry.client)
  return handler


@trace
def add_logging_context(tags, extra):
  from . import log_id
  tags = tags or {}
  extra = extra or {}
  ctx = log_id.ctx()
  tags.update(logid=ctx.get(defaults.LOG_KEY))
  extra.update(ctx)
  return tags, extra


def wrap_sentry(app):
  return RavenMiddleware(app, client=get_sentry().client)


class SentryClient(Client):
  """Subclass for more control over what is sent to sentry"""

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def capture(self, event_type, tags=None, extra=None, **kwargs):
    tags, extra = add_logging_context(tags, extra)
    super().capture(event_type, tags=tags, extra=extra, **kwargs)


class RavenMiddleware(middleware.Sentry):
  """Clear context before request starts"""

  def __call__(self, environ, start_response):
    self.client.context.clear()
    return super().__call__(environ, start_response)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Applicable to webstarts"""
import uuid

from structlog import get_logger

from . import defaults, trace

__author__ = "john"

logger = get_logger(__name__)


def err(v):
  ass = 1 + v
  return ass


def make():
  return str(uuid.uuid4())[:12].replace('-', '')


def ctx():
  return dict(logger._context._dict)


def find():
  return ctx().get(defaults.LOG_KEY)


def wrap_logid(app):
  @trace
  def wrap_logid_(environ, start_response):
    # maybe do not need to do .new if logid exists. in backend too.
    logid = environ.get(defaults.LOG_KEY, make())
    environ[defaults.LOG_KEY] = logid
    logger.new(logid=logid)

    @trace
    def wrap_logid_header(status, headers, exc_info=None):
      headers.append((defaults.LOG_KEY, logid))
      start_response(status, headers, exc_info)
    return app(environ, wrap_logid_header)
    # return ClosingIterator(app(environ, wrap_logid_header), lambda: None)
  return wrap_logid_



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Applicable to webstarts"""
from inspect import currentframe

__author__ = "john"

LOG_KEY = 'logid'
FORMAT = '%(levelname)s %(logid)s %(name)s %(message)s %(extras)s'
FORMAT_DEBUG = '%(levelname)-8s %(logid)-12s %(name)-30s %(message)-100s'
CACHE_KEY = lambda *a, **kws: id(currentframe().f_back.f_back.f_locals.get('wrapped'))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Applicable to webstarts"""
import logging
from io import StringIO

import colorama
from structlog import stdlib
from structlog.processors import KeyValueRenderer
from structlog.stdlib import _FixedFindCallerLogger

__author__ = "john"


class LogFactory(stdlib.LoggerFactory):
  """LoggerFactory for structlog"""

  def __init__(self, ignore_frame_names=None):
    super().__init__(ignore_frame_names=ignore_frame_names)
    logging.setLoggerClass(WebstartsLogger)


class WebstartsLogger(_FixedFindCallerLogger):
  """Root logger who will handle end result from structlog"""

  def _log(self, level, msg, *args, **kwargs):
    super()._log(level, msg, *args, **kwargs)

  def makeRecord(self, *args, **kwargs):
    return super().makeRecord(*args, **kwargs)


class WebstartsFormatter(logging.Formatter):
  """End logging formatter for all"""

  def __init__(self, fmt=None, proc=None, **kwargs):
    self.proc = proc
    super().__init__(fmt, **kwargs)

  def format(self, record):
    if self.proc:
      record.extras = self.proc.run(record)
    record.msg = f'"{record.msg}"'
    return super().format(record)


class ExtrasProcessor(KeyValueRenderer):
  """Replace ConsoleRenderer for final output """

  def run(self, record):
    ed = record.__dict__.get('ed', {})
    return ' '.join(k + '=' + self._repr(v) for k, v in self._ordered_items(ed))


class DevFormatter(logging.Formatter):
  """End logging formatter for all"""

  def __init__(self, proc=None, **kwargs):
    kwargs.pop('fmt', None)
    self.proc = proc
    super().__init__(fmt='%(message)s', **kwargs)

  def format(self, record):
    if self.proc:
      record.msg = self.proc.run(record)
    return super().format(record)


reset = colorama.Style.RESET_ALL
bright = colorama.Style.BRIGHT
dim = colorama.Style.DIM
red = colorama.Fore.RED
blue = colorama.Fore.BLUE
cyan = colorama.Fore.CYAN
magenta = colorama.Fore.MAGENTA
yellow = colorama.Fore.YELLOW
green = colorama.Fore.GREEN
rb = colorama.Back.RED
brblack = colorama.Fore.BLACK + bright


class DevRenderer(object):
  """Pretty shit"""

  NAME = blue + bright
  KEY = cyan
  VALUE = magenta
  LEVELS = dict(warning=colorama.Fore.LIGHTYELLOW_EX, info=green, debug=blue, notset=rb)
  LEVELS['critical'] = LEVELS['exception'] = LEVELS['error'] = red

  def __init__(self, name_pad=40, msg_pad=100):
    colorama.deinit()
    self.msg_pad = msg_pad
    self.name_pad = name_pad
    keys = self.LEVELS.keys()
    for k in keys:
      self.LEVELS[k] += bright

  def run(self, record):
    sio = StringIO()
    event_dict = record.__dict__.get('ed', {})
    namepad, msgpad, namecol, key, value = self.name_pad, self.msg_pad, self.NAME, self.KEY, self.VALUE
    name = record.name
    level = record.levelname.lower()
    levcol = self.LEVELS[level]
    logid = record.logid or 'None'
    msg = str(record.msg)
    if len(msg) > msgpad:
      msg = msg[:msgpad - 4] + '...'
    msg += '"'

    sio.write(f'[{levcol}{level:8}{reset}] {brblack}{logid:11}{reset} {namecol}{name:{namepad}}{reset} "{msg:{msgpad}}')

    stack = event_dict.pop('stack', None)
    exc = event_dict.pop('exception', None)

    atoms = list(event_dict.items())
    if atoms:
      atoms.sort(key=lambda e: str(e[0]))
      log_atom = ' '.join(f'{key}{k}{reset}{brblack}={reset}{value}{v}{reset}' for k, v in atoms)
      sio.write(f' {log_atom}')

    if stack is not None:
      sio.write("\n" + stack)
      if exc is not None:
        sio.write("\n\n" + "=" * 79 + "\n")
    if exc is not None:
      sio.write("\n" + exc)

    return sio.getvalue()


class StructFormatter(logging.Formatter):
  """End formatter maybe"""

  def __init__(self, processor, foreign_pre_chain=None, keep_exc_info=False, keep_stack_info=False, *args, **kwargs):
    self.processor = processor
    self.foreign_pre_chain = foreign_pre_chain
    self.keep_exc_info = keep_exc_info
    self.keep_stack_info = keep_stack_info
    fmt = kwargs.pop('fmt')
    if isinstance(processor, DevRenderer): fmt = '%(message)s'
    super().__init__(*args, fmt=fmt, **kwargs)

  # @trace
  def format(self, record):
    """Figure out what to do with exc_info"""

    if isinstance(record.msg, dict):
      ed = record.msg.copy()
    else:
      ed = {"event": record.getMessage(), "_record": record}
      record.args = ()
      # if record.exc_info:
      #   ed["exc_info"] = record.exc_info
      # if record.stack_info:
      #   ed["stack_info"] = record.stack_info
      # if not self.keep_exc_info:
      #   record.exc_text = None
      #   record.exc_info = None
      # if not self.keep_stack_info:
      #   record.stack_info = None
      for proc in self.foreign_pre_chain or ():
        ed = proc(None, record.name, ed)
      del ed['_record']
    record.msg = self.processor(record, record.name, ed)

    return super().format(record)



# coding=utf-8

from setuptools import find_packages, setup

setup(
  name='webstarts',
  version='0.0.8',
  license='BSD',
  description="Entry point for modern flask/gunicorn/sentry/celery web applications",
  long_description="Python entry point setup best practices for modern webapps",
  author='john',
  author_email='john@chaosdevs.com',
  url='https://github.com/j-walker23/webstarts',
  zip_safe=False,
  include_package_data=True,
  packages=find_packages(exclude=('wstests',)),
  python_requires='~=3.6',
  install_requires=[
    'celery>=4.0.2',
    'colorama>=0.3.9',
    'decorator>=4.0.11',
    'flask>=0.12.2',
    'flask-sqlalchemy>=2.2',
    'gevent>=1.2.2',
    'gunicorn>=19.7.1',
    'kids.cache>=0.0.4',
    'raven[flask]>=6.1.0',
    'requests>=2.18.1',
    'sqlalchemy>=1.1.11',
    'structlog>=17.2.0',
  ],
  entry_points='''
    [console_scripts]
    webstarts=webstarts:entry
  ''',
  classifiers=[
    'Intended Audience :: Developers',
    'Intended Audience :: System Administrators',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 3.6',
    'Topic :: Software Development',
  ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Applicable to webstarts"""
import threading

import requests
from decorator import decorator

from . import defaults, log_id

__author__ = "john"
__all__ = ['req_session']


class Local(threading.local):
  def __init__(self):
    super().__init__()
    self.cache = {}


_local = Local()
_key = requests.Session


def req_session() -> requests.Session:
  """Thread local request sessions"""

  if _key not in _local.cache:
    if not getattr(_key.prepare_request, '_webstarts', None):
      _key.prepare_request = _pre_hook(_key.prepare_request)

    _local.cache[_key] = _key()

  return _local.cache[_key]


def wrap_session(app):
  """Add req_session as middleware"""

  session = req_session()

  def wrap_session_(environ, start_response):
    environ.update(requests=session)
    return app(environ, start_response)
  return wrap_session_


@decorator
def _pre_hook(f, *args):
  prepared = f(*args)
  if defaults.LOG_KEY not in prepared.headers:
    prepared.headers[defaults.LOG_KEY] = log_id.find()
  prepared._webstarts = True
  return prepared



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Applicable to webstarts"""
import json
import os

from decorator import decorator

__author__ = "john"


def get_conf(env):
  """Right now just checks if it is debug mode"""
  conf = json.loads(env.get('WSCONF', '{}'))
  debug = True if env.get('WSDEV') == '1' else False
  conf.update(debug=debug)
  return conf


def configure(env=os.environ):
  from . import wlog, backend
  conf = get_conf(env)
  wlog.setup(conf)
  backend.setup_celery(conf)
  return conf


def entry():
  """Run eventlet


  - structlog
    Check out colorama
    Like queries always do log = log.new(y=23)
    logging.getLogger('foo').addHandler(logging.NullHandler()) should be used by libraries
  - GoogleCloud logging - when the shit gets fixed
    I could add the handler to logging.getLogger('socialclime') since it will be the parent of all app loggers.
    Now that i finally understand logging.

  """
  from gevent import monkey
  monkey.patch_all(subprocess=True)

  conf = configure()

  from .gunicorn import WebstartsApp

  app = WebstartsApp('%(prog)s [OPTIONS] [APP_MODULE]', conf['debug'])
  return app.run()


@decorator
def trace(f, *args, **kwargs):
  import colorama
  import inspect
  reset = colorama.Style.RESET_ALL
  brblack = colorama.Fore.BLACK + colorama.Style.BRIGHT
  red = colorama.Fore.LIGHTRED_EX + colorama.Style.BRIGHT
  # full = inspect.getcallargs(f, *args, **kwargs)
  # full = dict(full)
  # aas, kws = full.pop('args', None), full.pop('kwargs', None)
  # print(f'{brblack}{name:50}{reset} {brblack}Named={reset}{full!r:200} {brblack}Args={reset}{aas} {brblack}Kwargs={reset}{kws}')

  from . import log_id
  name = [i for i in inspect.getmembers(f, lambda mem: isinstance(mem, str)) if i[0] == '__qualname__'][0][1]
  before = log_id.find()
  ret = f(*args, **kwargs)
  after = log_id.find()
  # print(f'{brblack}{name:100}{reset} - {red}{before}{reset}->{brblack}{after}{reset}')
  return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Applicable to webstarts"""

import logging
import sys
from typing import Any, Dict

from kids.cache import cache
from structlog import configure_once, processors, stdlib, threadlocal

from . import defaults, types

__author__ = "john"


def setup(conf: Dict[str, Any]):
  """Setup logging first so nothing shits on it

  - Checkout
  """
  logging.setLoggerClass(types.WebstartsLogger)
  logging.getLogger().setLevel(logging.NOTSET)
  set_record_factory()

  formatter = get_formatter(conf)
  handler = get_handler()
  set_handler(logging.DEBUG, handler, formatter)

  setup_struct()
  shush_loggers()

  logger = logging.getLogger(__name__)
  logger.info('Configured Logging')

  from .wflask import sentry_handler
  set_handler(logging.ERROR, sentry_handler(), formatter)


@cache(key=defaults.CACHE_KEY)
def get_formatter(conf: Dict[str, str] = None) -> types.WebstartsFormatter:
  conf = conf or {}
  fmt, proc = conf.get('fmt', None), None
  if conf['debug']:
    formatter = types.DevFormatter(fmt=fmt or defaults.FORMAT_DEBUG, proc=types.DevRenderer())
  else:
    formatter = types.WebstartsFormatter(fmt=fmt or defaults.FORMAT, proc=types.ExtrasProcessor())
  return formatter


def set_handler(level: int, handler: logging.Handler, formatter: logging.Formatter = None) -> None:
  if formatter:
    handler.setFormatter(formatter)
  handler.setLevel(level)
  logging.getLogger().addHandler(handler)


def setup_struct() -> None:
  """Configure the shit"""
  configure_once(
    processors=[
      stdlib.filter_by_level,
      stdlib.PositionalArgumentsFormatter(),
      processors.UnicodeDecoder(),
      render_to_log_kwargs,
    ],
    context_class=threadlocal.wrap_dict(dict),
    logger_factory=types.LogFactory(),
    wrapper_class=stdlib.BoundLogger,
    cache_logger_on_first_use=True,
  )


@cache(key=defaults.CACHE_KEY)
def get_handler() -> logging.StreamHandler:
  return logging.StreamHandler(sys.stdout)


def set_record_factory() -> None:
  old_factory = logging.getLogRecordFactory()
  # noinspection PyMissingTypeHints
  def record_factory(*args, **kwargs):
    from webstarts import log_id
    record = old_factory(*args, **kwargs)
    record.logid = log_id.find()
    record.extras = ''
    return record
  logging.setLogRecordFactory(record_factory)


def render_to_log_kwargs(_, __, event_dict) -> Dict:
  extra = event_dict.pop('extra', {})
  exc = event_dict.pop('exc_info', None)
  stack_info = event_dict.pop('stack_info', False)
  event_dict.pop('logid', None)
  ret = dict(msg=event_dict.pop("event", ''), extra={'ed': event_dict, **extra}, exc_info=exc, stack_info=stack_info)
  return ret


def shush_loggers():
  names = [
    'passlib'
  ]
  for n in names:
    logging.getLogger(n).setLevel(logging.WARN)
